#include "Log.h"
#include "arch/CPU.h"
#include "arch/Serial.h"
#include "arch/Timer.h"
#include "boot/Init.h"
#include "config.h"
#include "memory/MemoryManager.h"
#include "thread/Scheduler.h"
#include <luna/Result.h>
#include <luna/Units.h>

void async_thread()
{
    while (true)
    {
        Thread* current = Scheduler::current();
        kinfoln("Ticks: %lu, %lu user, %lu kernel, %lu idle", current->ticks, current->ticks_in_user,
                current->ticks_in_kernel, Scheduler::idle()->ticks);

        CPU::wait_for_interrupt();

        kernel_sleep(1000);
    }
}

Result<void> init()
{
    kinfoln("Starting Moon %s", MOON_VERSION);

    kinfoln("Current platform: %s", CPU::platform_string());

    kinfoln("Current processor: %s", CPU::identify().value_or("(unknown)"));

    Timer::init();

    char buffer[64];
    to_dynamic_unit(MemoryManager::total(), buffer, sizeof(buffer));
    kinfoln("Total memory:    %s", buffer);
    to_dynamic_unit(MemoryManager::free(), buffer, sizeof(buffer));
    kinfoln("Free memory:     %s", buffer);
    to_dynamic_unit(MemoryManager::used(), buffer, sizeof(buffer));
    kinfoln("Used memory:     %s", buffer);
    to_dynamic_unit(MemoryManager::reserved(), buffer, sizeof(buffer));
    kinfoln("Reserved memory: %s", buffer);

    Scheduler::init();

    TRY(Scheduler::new_kernel_thread(async_thread));

    CPU::platform_finish_init();

    CPU::enable_interrupts();

    return {};
}

extern "C" [[noreturn]] void _start()
{
    Init::check_magic();
    Init::early_init();
    auto rc = init();
    if (rc.has_error()) kerrorln("Runtime error: %s", rc.error_string());
    CPU::idle_loop();
}